package eu.ztsh.wymiana.service;

import eu.ztsh.wymiana.EntityCreator;
import eu.ztsh.wymiana.RepositoryBasedTest;
import eu.ztsh.wymiana.config.CurrencyProperties;
import eu.ztsh.wymiana.data.repository.UserRepository;
import eu.ztsh.wymiana.exception.InsufficientFundsException;
import eu.ztsh.wymiana.exception.UserNotFoundException;
import eu.ztsh.wymiana.model.Symbol;
import eu.ztsh.wymiana.validation.InstanceValidator;
import eu.ztsh.wymiana.validation.ValidationFailedException;
import jakarta.transaction.Transactional;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;

import java.math.BigDecimal;
import java.util.Objects;
import java.util.stream.Stream;

import static eu.ztsh.wymiana.EntityCreator.Constants.*;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

class CurrencyServiceTest extends RepositoryBasedTest {

    private final CurrencyService currencyService;

    @Autowired
    public CurrencyServiceTest(UserRepository userRepository, InstanceValidator instanceValidator,
                               CurrencyProperties currencyProperties) {
        super(userRepository);
        var nbp = Mockito.mock(NbpService.class);
        Mockito.when(nbp.getSellRate(Symbol.USD)).thenReturn(SELL_RATE);
        Mockito.when(nbp.getBuyRate(Symbol.USD)).thenReturn(BUY_RATE);
        currencyService = new CurrencyService(new UserService(userRepository, instanceValidator, currencyProperties),
            nbp, instanceValidator);
    }

    @Transactional
    @Test
    void plnToUsdToSellSuccessTest() {
        var entity = EntityCreator.userEntity().build();
        userRepository.save(entity);
        var result = currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toSell(PLN)
            .build());
        assertThat(result.currencies())
            .matches(map -> map.get(Symbol.PLN).amount().signum() == 0 && Objects.equals(map.get(Symbol.USD).amount(), USD_BUY), "USD 5.08");
        var expected = EntityCreator.userEntity().pln(0).usd(USD_BUY).build();
        expect(expected);
    }

    @Transactional
    @Test
    void plnToUsdToBuySuccessTest() {
        var entity = EntityCreator.userEntity().build();
        userRepository.save(entity);
        var result = currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toBuy(USD_BUY)
            .build());
        assertThat(result.currencies())
            .matches(map -> map.get(Symbol.PLN).amount().signum() == 0 && Objects.equals(map.get(Symbol.USD).amount(), USD_BUY));
        var expected = EntityCreator.userEntity().pln(0).usd(USD_BUY).build();
        expect(expected);
    }

    @Transactional
    @Test
    void usdToPlnToSellSuccessTest() {
        var entity = EntityCreator.userEntity().pln(-1).usd(USD_SELL).build();
        userRepository.save(entity);
        var result = currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.USD)
            .to(Symbol.PLN)
            .toSell(USD_SELL)
            .build());
        assertThat(result.currencies())
            .matches(map -> Objects.equals(map.get(Symbol.PLN).amount(), PLN) && map.get(Symbol.USD).amount().signum() == 0, "PLN 20.10");
        var expected = EntityCreator.userEntity().pln(PLN).usd(0).build();
        expect(expected);
    }

    @Transactional
    @Test
    void usdToPlnToBuySuccessTest() {
        var entity = EntityCreator.userEntity().pln(-1).usd(USD_SELL).build();
        userRepository.save(entity);
        var result = currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.USD)
            .to(Symbol.PLN)
            .toBuy(PLN)
            .build());
        assertThat(result.currencies())
            .matches(map -> Objects.equals(map.get(Symbol.PLN).amount(), PLN) && map.get(Symbol.USD).amount().signum() == 0, "PLN 20.10");
        var expected = EntityCreator.userEntity().pln(PLN).usd(0).build();
        expect(expected);
    }

    @Transactional
    @Test
    void usdToPlnNoUsdCurrencyTest() {
        var entity = EntityCreator.userEntity().build();
        userRepository.save(entity);
        var request = EntityCreator.exchangeRequest()
            .from(Symbol.USD)
            .to(Symbol.PLN)
            .toBuy(PLN)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(request))
            .isInstanceOf(InsufficientFundsException.class);
    }

    @Transactional
    @Test
    void doubleExchangeTest() {
        var initialValue = BigDecimal.valueOf(100);
        var entity = EntityCreator.userEntity().pln(initialValue).build();
        userRepository.save(entity);
        currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toBuy(USD_BUY)
            .build());
        currencyService.exchange(EntityCreator.exchangeRequest()
            .from(Symbol.USD)
            .to(Symbol.PLN)
            .toSell(USD_BUY)
            .build());
        var resultOptional = userRepository.findById(entity.getPesel());
        assertThat(resultOptional)
            .isNotEmpty();
        var resultEntity = resultOptional.get();
        assertThat(resultEntity.getCurrencies()
            .stream()
            .filter(c -> c.getSymbol().equals(Symbol.PLN))
            .findFirst()).isNotEmpty().get().matches(currencyEntity -> currencyEntity.getAmount().compareTo(initialValue) < 0);
    }

    @Transactional
    @Test
    void insufficientFundsTest() {
        var entity = EntityCreator.userEntity().build();
        userRepository.save(entity);
        var request = EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toBuy(PLN)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(request))
            .isInstanceOf(InsufficientFundsException.class);
    }

    @DisplayName("Invalid PESEL value")
    @ParameterizedTest
    @MethodSource
    void invalidPeselTest(String pesel) {
        var entity = EntityCreator.exchangeRequest()
            .pesel(pesel)
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toSell(USD_SELL)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(entity))
            .isInstanceOf(ValidationFailedException.class)
            .hasMessageContaining("PESEL");
    }

    @Test
    void notExistingUserTest() {
        var entity = EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .to(Symbol.USD)
            .toSell(USD_SELL)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(entity))
            .isInstanceOf(UserNotFoundException.class);
    }

    @Test
    @DisplayName("Empty 'from' value")
    void emptyFromTest() {
        var entity = EntityCreator.exchangeRequest()
            .to(Symbol.USD)
            .toSell(USD_SELL)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(entity))
            .isInstanceOf(ValidationFailedException.class)
            .hasMessageContaining("null");
    }

    @Test
    @DisplayName("Empty 'to' value")
    void emptyToTest() {
        var entity = EntityCreator.exchangeRequest()
            .from(Symbol.PLN)
            .toSell(USD_SELL)
            .build();
        assertThatThrownBy(() -> currencyService.exchange(entity))
            .isInstanceOf(ValidationFailedException.class)
            .hasMessageContaining("null");
    }

    private static Stream<String> invalidPeselTest() {
        return Stream.of("INVALID", INVALID_PESEL);
    }

}
