package eu.ztsh.wymiana.web;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import eu.ztsh.wymiana.EntityCreator;
import eu.ztsh.wymiana.WireMockExtension;
import eu.ztsh.wymiana.model.Symbol;
import eu.ztsh.wymiana.util.UserMapper;
import org.junit.jupiter.api.ClassOrderer;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.MethodOrderer;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestClassOrder;
import org.junit.jupiter.api.TestMethodOrder;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.web.reactive.server.WebTestClient;

import java.math.BigDecimal;
import java.time.DayOfWeek;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAdjusters;

import static eu.ztsh.wymiana.EntityCreator.Constants.*;
import static org.assertj.core.api.Assertions.assertThat;

/**
 * Integration test suite.
 * Contrary to the principle of test independence, tests are dependent on one another to create continuous suite.
 */
@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
@ActiveProfiles("it")
@ExtendWith(WireMockExtension.class)
@TestMethodOrder(MethodOrderer.DisplayName.class)
@TestClassOrder(ClassOrderer.DisplayName.class)
class ApplicationIntegrationTests {

    private final WebTestClient webTestClient;
    private final ObjectMapper objectMapper;

    @Autowired
    public ApplicationIntegrationTests(WebTestClient webTestClient) {
        this.webTestClient = webTestClient;
        objectMapper = new ObjectMapper();
    }

    private String asJson(Object object) {
        try {
            return objectMapper.writeValueAsString(object);
        } catch (JsonProcessingException e) {
            throw new IllegalStateException(e);
        }
    }

    @Nested
    @TestMethodOrder(MethodOrderer.DisplayName.class)
    @DisplayName("01: Context")
    class ContextTests {

        @Test
        @DisplayName("01.1: Load context")
        void contextLoads() {
            assertThat(webTestClient).isNotNull();
            webTestClient.get().uri("/actuator/health").exchange().expectBody().json("{\"status\":\"UP\"}");
        }

    }

    @Nested
    @TestMethodOrder(MethodOrderer.DisplayName.class)
    @DisplayName("02: User")
    class UserTests {

        private static final String endpoint = "/api/user";

        @Test
        @DisplayName("02.1: Create valid user")
        void createUserTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.userRequest(100D).build())
                .exchange()
                .expectStatus().isNoContent();
        }

        @Test
        @DisplayName("02.2: Try to create invalid user")
        void createInvalidUserTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.userRequest().pesel(INVALID_PESEL).build())
                .exchange()
                .expectStatus()
                .isBadRequest();
        }

        @Test
        @DisplayName("02.3: Try to create duplicated user")
        void createDuplicatedUserTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.userRequest().build())
                .exchange()
                .expectStatus().isEqualTo(409)
                .expectBody(String.class).isEqualTo("User with PESEL %s already exists".formatted(PESEL));
        }

        @Test
        @DisplayName("02.4: Get valid user")
        void getValidUserTest() {
            webTestClient.get()
                .uri(endpoint.concat("/").concat(PESEL))
                .exchange()
                .expectStatus().isOk()
                .expectBody().json(asJson(UserMapper.pojoToResponse(UserMapper.entityToPojo(EntityCreator.userEntity().pln(100).build()))));
        }

        @Test
        @DisplayName("02.5: Try to get non-existing user")
        void getNonExistingUserTest() {
            webTestClient.get()
                .uri(endpoint.concat("/").concat(ANOTHER_PESEL))
                .exchange()
                .expectStatus().isNotFound();
        }

        @Test
        @DisplayName("02.6: Get user by incorrect PESEL")
        void getIncorrectPeselUserTest() {
            webTestClient.get()
                .uri(endpoint.concat("/").concat(INVALID_PESEL))
                .exchange()
                .expectStatus().isBadRequest();
        }

    }

    @Nested
    @TestMethodOrder(MethodOrderer.DisplayName.class)
    @DisplayName("03: Exchange")
    class ExchangeTests {

        private static final String URI_PATTERN = "/api/exchangerates/rates/c/usd/%s/";
        private final DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd");
        private static final String endpoint = "/api/exchange";

        @Test
        @DisplayName("03.1: Try to perform invalid money exchange: no data")
        void noNbpDataTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.exchangeRequest()
                    .from(Symbol.PLN)
                    .to(Symbol.USD)
                    .toSell(PLN)
                    .build())
                .exchange()
                .expectStatus().isEqualTo(500);
        }

        @Test
        @DisplayName("03.2: Perform valid money exchange")
        void exchangeTest() {
            var date = getTodayOrLastFriday();
            WireMockExtension.response(
                URI_PATTERN.formatted(date),
                200,
                asJson(EntityCreator.rates(date))
            );
            var expected = asJson(EntityCreator.user(BigDecimal.valueOf(100).subtract(PLN), USD_BUY));
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.exchangeRequest()
                    .from(Symbol.PLN)
                    .to(Symbol.USD)
                    .toSell(PLN)
                    .build())
                .exchange()
                .expectStatus().isOk()
                .expectBody().json(expected);
            WireMockExtension.verifyGet(1, URI_PATTERN.formatted(date));
        }

        @Test
        @DisplayName("03.3: Try to perform invalid money exchange: not existing user")
        void exchangeNotExistingUserTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.exchangeRequest()
                    .pesel(ANOTHER_PESEL)
                    .from(Symbol.PLN)
                    .to(Symbol.USD)
                    .toSell(PLN)
                    .build())
                .exchange()
                .expectStatus().isNotFound()
                .expectBody(String.class).isEqualTo("User with PESEL %s not found".formatted(ANOTHER_PESEL));
        }

        @Test
        @DisplayName("03.4: Try to perform invalid money exchange: invalid PESEL")
        void invalidPeselTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.exchangeRequest()
                    .pesel(INVALID_PESEL)
                    .from(Symbol.PLN)
                    .to(Symbol.USD)
                    .toSell(PLN)
                    .build())
                .exchange()
                .expectStatus().isBadRequest();
        }

        @Test
        @DisplayName("03.5: Try to perform invalid money exchange: insufficient funds")
        void insufficientFundsTest() {
            webTestClient.post()
                .uri(endpoint)
                .bodyValue(EntityCreator.exchangeRequest()
                    .from(Symbol.USD)
                    .to(Symbol.PLN)
                    .toBuy(PLN)
                    .build())
                .exchange()
                .expectStatus().isBadRequest();
        }

        @Test
        @DisplayName("03.6: Perform valid money exchange with lower case currency symbols")
        void ignoreCaseTest() {
            var date = getTodayOrLastFriday();
            var expected = asJson(EntityCreator.user(BigDecimal.valueOf(100).subtract(PLN.multiply(BigDecimal.TWO)), USD_BUY.multiply(BigDecimal.TWO)));
            webTestClient.post()
                .uri(endpoint)
                .header("Content-Type", "application/json")
                .bodyValue(asJson(EntityCreator.exchangeRequest()
                    .from(Symbol.PLN)
                    .to(Symbol.USD)
                    .toSell(PLN)
                    .build()).replace("USD", "usd").replace("PLN", "pln"))
                .exchange()
                .expectStatus().isOk()
                .expectBody().json(expected);
            // cache
            WireMockExtension.verifyGet(0, URI_PATTERN.formatted(date));
        }

        private String getTodayOrLastFriday() {
            var today = LocalDate.now();
            if (today.getDayOfWeek() == DayOfWeek.SATURDAY
                || today.getDayOfWeek() == DayOfWeek.SUNDAY) {
                today = today.with(TemporalAdjusters.previous(DayOfWeek.FRIDAY));
            }
            return today.format(dtf);
        }

    }

}
